@@ -1,6 +1,6 @@
 #![cfg(all(feature = "macros", not(PyPy)))]
 
-use pyo3::prelude::*;
+use pyo3::{prelude::*, types::PySuper};
 
 #[pyclass(subclass)]
 struct BaseClass {
@@ -29,22 +29,29 @@ impl SubClass {
         (SubClass {}, BaseClass::new())
     }
 
-    fn method(self_: &PyCell<Self>) -> PyResult<&PyAny> {
+    fn method<'py>(self_: &Bound<'py, Self>) -> PyResult<Bound<'py, PyAny>> {
         let super_ = self_.py_super()?;
         super_.call_method("method", (), None)
     }
+
+    fn method_super_new<'py>(self_: &Bound<'py, Self>) -> PyResult<Bound<'py, PyAny>> {
+        #[cfg_attr(not(feature = "gil-refs"), allow(deprecated))]
+        let super_ = PySuper::new_bound(&self_.get_type(), self_)?;
+        super_.call_method("method", (), None)
+    }
 }
 
 #[test]
 fn test_call_super_method() {
     Python::with_gil(|py| {
-        let cls = py.get_type::<SubClass>();
+        let cls = py.get_type_bound::<SubClass>();
         pyo3::py_run!(
             py,
             cls,
             r#"
         obj = cls()
         assert obj.method() == 10
+        assert obj.method_super_new() == 10
     "#
         )
     });
